# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

__all__=['packageManager','display','services','utils']

#shortcut to logging
import utils
import utils.pdLogging
logger = utils.pdLogging.getPixiedustLogger()
getLogger = utils.pdLogging.getLogger

#shortcut to packageManager
import packageManager
printAllPackages=packageManager.printAllPackages
installPackage=packageManager.installPackage
uninstallPackage=packageManager.uninstallPackage

import display
import services

#automated import into the user namespace
try:
    get_ipython().user_ns["display"]=display.display

    #javaBridge and scalaBridge only work in the driver, not an executor
    from utils.javaBridge import *
    from utils.scalaBridge import *
except NameError:
    #IPython not available we must be in a spark executor
    pass


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
'''
Provide easy access to the Shell user variables
Sample use:
    ShellAccess.sc => access to the SparkContext
    ShellAccess["sqlContext"] => access the SQLContext
    ShellAccess.myVar = "Hello" => set a new variable called myVar in the user namespace
'''
class ShellAccess(object):
    __metaclass__= type("",(type,),{
        "__getitem__":lambda cls, key: get_ipython().user_ns.get(key),
        "__setitem__":lambda cls, key,val: get_ipython().user_ns.update({key:val}),
        "__getattr__":lambda cls, key: get_ipython().user_ns.get(key),
        "__setattr__":lambda cls, key, val: get_ipython().user_ns.update({key:val}),
        "__iter__": lambda cls: iter(get_ipython().user_ns.keys())
    })

    @staticmethod
    def update(**kwargs):
        for key,val in kwargs.iteritems():
            Configuration[key]=val



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import logging
from logging.handlers import MemoryHandler
from IPython.core.magic import (Magics, magics_class, line_magic)

from collections import deque
logMessageBufferSize = 200
logMessages=deque([], logMessageBufferSize)

class PixiedDustLoggingHandler(logging.Handler):
    def emit(self, record):
        logMessages.append((record.levelno, self.format(record)))

#init pixiedust logging
pixiedustHandler = PixiedDustLoggingHandler()
pixiedustHandler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
memHandler = MemoryHandler(1, target=pixiedustHandler)
memHandler.setLevel(logging.DEBUG)

pixiedustLogger = logging.getLogger("PixieDust")
pixiedustLogger.addHandler(memHandler)
pixiedustLogger.setLevel(logging.DEBUG)

@magics_class
class PixiedustLoggingMagics(Magics):
    @line_magic
    def pixiedustLog(self, arg_s):
        try:
            opts,args = self.parse_options( arg_s, "l:f:m:")
            level = logging.getLevelName( opts.get("l", "INFO").upper() )
            if not isinstance(level, int):
                level = logging.getLevelName("INFO")
            lookup = opts.get("f", None)
            maxLines = int(opts.get("m", logMessageBufferSize))
            start = len(logMessages) - maxLines - 1
            for i, (l,record) in enumerate(logMessages):
                if i > start and l >= level and (lookup is None or lookup in record): 
                    print(record)
        except:
            print("""Usage Error: -l logLevel -f lookupString -m 20
                -l: filter by log level e.g  'CRITICAL':'FATAL': 'ERROR': 'WARNING':'INFO':'DEBUG'
                -f: filter message that contain given string e.g. "Exception" 
                -m: max number of log messages returned
            """)
            raise

try:
    get_ipython().register_magics(PixiedustLoggingMagics)
except NameError:
    #IPython not available we must be in a spark executor
    pass

def getPixiedustLogger():
    return pixiedustLogger

loggerDict={}
def getLogger(loggerName):
    logger = loggerDict.get(loggerName)
    if logger is None:
        logger = logging.getLogger(loggerName)
        logger.addHandler(memHandler)
        logger.setLevel(logging.DEBUG)
        logger.propagate=0
        loggerDict[loggerName]=logger
    return logger


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import Display
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
import yaml

class GraphDisplay(Display):
    def doRender(self,handlerId):
        g=self.entity
        
        if ( handlerId == "nodeLinkGraph"):
            import json
            ar = g.edges.select("src","dst").map(lambda (s,d): (s,[d]))\
                .reduceByKey(lambda d1,d2: d1+d2).map(lambda (src, arTargets): (src, list(set(arTargets))))\
                .collect()

            dic = {item[0] : item[1] for item in ar}
            limitLevel = self.options.get("limitLevel", 5)
            limitChildren = self.options.get("limitChildren", 10)
            def expand(values, visited,level):
                results=[]
                if values is not None and level < limitLevel:
                    for v in values:
                        if v not in visited and len(results)<limitChildren:
                            visited[v]=True
                            results.append({ "name": str(v), "children": {}})
                    for item in results:
                        nextVisited = {}
                        nextVisited.update(visited)
                        item["children"]=expand(dic.get(item["name"]), nextVisited, level+1)
                return results
            
            root = self.options.get("root")
            rootNode = ar[0]
            if root:
                def findRoot(ar):
                    for a in ar:
                        if a[0]==root:
                            return a
                rootNode = findRoot(ar)
            
            if not rootNode:
                self._addHTML("<p>Can't find the airport</p>");
                return;

            res = { "name": str(rootNode[0]), "children":expand(dic[ar[0][0]], {rootNode[0]:True}, 1)}
            tree = json.dumps(res)

            #if user specified root, then only send back the json tree
            if root:
                self.addProfilingTime = False
                print(tree)
                return

            self._addScriptElement("https://d3js.org/d3.v3.js", checkJSVar="d3", callback=self.renderTemplate("nodeLinkGraph.js", root=tree))
            self._addHTMLTemplate("nodeLinkGraph.html", root=tree, res=res)
        else:
            graphNodesJson="{"
            for r in g.vertices.map(lambda row: """"{0}":{{"id":"{0}","name":"{1}","latitude":{2},"longitude":{3}}}"""
                .format(row.id, row.name.encode("ascii","ignore"),0.0 if row.latitude is None else row.latitude,0.0 if row.longitude is None else row.longitude)).collect():
                graphNodesJson+=("," if len(graphNodesJson)>1 else "") + str(r)
            graphNodesJson+="}"        
            graphLinksJson=str(g.edges.select("src","dst").groupBy("src","dst").agg(F.count("src").alias("count")).toJSON().map(lambda j: yaml.safe_load(j)).collect())
    
            self._addScriptElement("https://d3js.org/d3.v3.js", checkJSVar="d3")
            self._addScriptElement("https://mbostock.github.io/d3/talk/20111116/d3/d3.geo.js")
            self._addScriptElement(
                "https://mbostock.github.io/d3/talk/20111116/d3/d3.geom.js", 
                callback= self.renderTemplate("graphMap.js", graphNodesJson=graphNodesJson, graphLinksJson=graphLinksJson)
            )        
            self._addHTMLTemplate("graphMap.html")



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pyspark import SparkContext
from pyspark.sql import SQLContext, DataFrame
import py4j.java_gateway
import sys
import pixiedust
from py4j.java_gateway import *

myLogger = pixiedust.getLogger(__name__)

"""
class used to redirect Java JVM System output to Python notebook
It implements a Java interface that will be used as a callback from the java side
"""
class PixiedustOutput(object):
    def printOutput(self, s):
        print(s)

    def sendChannel(self, channel, data):
        self.printOutput(channel + " : " + data)
        
    class Java:
        implements = ["com.ibm.pixiedust.PixiedustOutputListener"]

"""
Helper class for making it easier to call Java code from Python
"""
sc = SparkContext.getOrCreate()
class JavaWrapper(object): 
    def __init__(self, objectIdentifier, captureOutput=False, outputChannel = None, outputReceiverClassName = None):
        if isinstance(objectIdentifier, str if sys.version >= '3' else basestring):
            self.jHandle = self._getJavaHandle(objectIdentifier)
        else:
            self.jHandle=objectIdentifier
        self.outputChannel = self.createClass(outputChannel) or PixiedustOutput()
        self.outputReceiver = JavaWrapper(outputReceiverClassName) if outputReceiverClassName is not None else None
        self.captureOutput(captureOutput)

    def createClass(self, className):
        if className is None or className == False:
            return None
        #extract module
        index = className.rfind(".")
        moduleName = None
        if index > 0:
            moduleName = className[:index]
            className = className[index+1:]

        __import__(moduleName)
        return getattr(sys.modules[moduleName], className)()
        
    def captureOutput(self, doIt):
        if doIt:
            if self.outputReceiver and self.outputChannel and self.outputReceiver.hasMethod("setChannelListener", sc._jvm.java.lang.Class.forName("com.ibm.pixiedust.PixiedustOutputListener")):
                self.outputReceiver.setChannelListener(self.outputChannel)
            else:
                pixiedustOutputSink = JavaWrapper("com.ibm.pixiedust.PixiedustOutputStream").jHandle(self.outputChannel)
                JavaWrapper("scala.Console").setOut(pixiedustOutputSink)

            #TODO: find a way to redirect system.out output for the current execution of the call
            #JavaWrapper("java.lang.System").setOut(pixiedustOutputSink)
            try:
                gw = sc._gateway
                if "_callback_server" not in gw.__dict__ or gw._callback_server is None:
                    gw.callback_server_parameters.eager_load = True
                    gw.callback_server_parameters.daemonize = True
                    gw.callback_server_parameters.daemonize_connections = True
                    gw.callback_server_parameters.port = 0
                    gw.start_callback_server(gw.callback_server_parameters)
                    gw._callback_server.port = gw._callback_server.server_socket.getsockname()[1]
                    jgws = JavaObject("GATEWAY_SERVER", gw._gateway_client)
                    gw.jvm.org.apache.spark.streaming.api.python.PythonDStream.updatePythonGatewayPort(jgws, gw._callback_server.port)
            except Exception, e:
                print("Error {0}".format(e))

    def __getattr__(self, name):
        o = self.jHandle.__getattr__(name)
        return o
        
    def _getJavaHandle(self,fqName):
        parts = fqName.split(".")
        handle = sc._jvm
        for package in parts:
            handle = handle.__getattr__(package)
        try:
            handle = handle.__getattr__("MODULE$")
        except:
            handle = handle
        return handle

    def hasMethod(self, methodName, *args):
        return self.getMethod(methodName, *args) != None

    def getMethods(self):
        return self.jHandle.getClass().getMethods()

    def getMethod(self, methodName, *args):
        argLen = len(args)
        jMethodParams = None if argLen == 0 else sc._gateway.new_array(sc._jvm.Class, argLen )
        for i,arg in enumerate(args):
            jMethodParams[i] = arg if arg.getClass().getName() == "java.lang.Class" else arg.getClass()

        try:
            return self.jHandle.getClass().getMethod(methodName, jMethodParams)
        except:
            return None

    '''
    use direct Java Reflection to run a method. This is needed because in case of dynamically loaded classes, Py$j automatic reflexion doesn't
    work right
    '''
    def callMethod(self, methodName, *args):
        myLogger.debug("Calling callMethod: methodName: {0}".format(methodName))
        argLen = len(args)
        jMethodParams = None if argLen == 0 else sc._gateway.new_array(sc._jvm.Class, argLen )
        jMethodArgs = None if argLen == 0 else sc._gateway.new_array(sc._jvm.Object, argLen )
        for i,arg in enumerate(args):
            jMethodParams[i] = (arg if arg.__class__.__name__ == "JavaClass" else arg.getClass())
            jMethodArgs[i] = arg
        #find the method and invoke it
        for m in self.jHandle.getClass().getMethods():
            if m.getName() == methodName and len(m.getParameterTypes()) == argLen:
                myLogger.debug("Found method with Name {0}".format(methodName))
                #Check the arguments type match
                match=True
                if argLen > 0:
                    for m1,m2 in zip( m.getParameterTypes(), jMethodParams ):
                        if not m1.isAssignableFrom(m2):
                            myLogger.debug("Found method {0} with arguments that are not matching: {1} => {2}".format(methodName, m1, m2))
                            match = False
                            break;
                if match:
                    return m.invoke(self.jHandle, jMethodArgs)
        
        raise ValueError("Method {0} that matches the given arguments not found".format(methodName) )
        #return self.jHandle.getClass().getMethod(methodName, jMethodParams).invoke(self.jHandle, jMethodArgs)

"""
Misc helper functions
"""
def pd_getJavaSparkContext():
    return sc._jsc.sc()

def pd_convertFromJava(entity):
    if ( entity.__class__.__name__ == "JavaMember"):
        entity = entity()
    if entity.__class__.__name__ == "JavaObject":
        javaClassName = entity.getClass().getName()
        if javaClassName == "org.apache.spark.sql.DataFrame":
            return DataFrame(entity, SQLContext(sc, entity.sqlContext()))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display.display import Display
from .serviceManager import *
import time
import requests
import json

CLOUDANT_CONN_TYPE = "cloudant"

class StashCloudantHandler(Display):
    def doRender(self, handlerId):
        entity=self.entity

        dbName = self.options.get("dbName", "dataframe-"+time.strftime('%Y%m%d-%H%M%S'))
        connectionName=self.options.get("connection")
        if connectionName is None:
            self._addHTMLTemplate("stashCloudant.html",dbName=dbName,connections=getConnections(CLOUDANT_CONN_TYPE))
        else:
            #first create the stash db
            connection = getConnection(CLOUDANT_CONN_TYPE, connectionName)
            if not connection:
                raise Exception("Unable to resolve connection {0}".format(connectionName))
            payload = json.loads( connection["PAYLOAD"])
            credentials=payload["credentials"]
            r = requests.put( credentials["url"] + "/" + dbName )
            if ( r.status_code != 200 and r.status_code != 201 ):
                print("Unable to create db ({0}) for connection ({1}): {2}".format(dbName, connectionName, str(r.content)))
            else:
                self.entity.write.format("com.cloudant.spark")\
                    .option("cloudant.host", credentials["host"])\
                    .option("cloudant.username",credentials["username"])\
                    .option("cloudant.password",credentials["password"])\
                    .option("createDBOnSave","true")\
                    .save(dbName)
                print("""Successfully stashed your data: <a target='_blank' href='{0}/{1}'>{1}</a>""".format(credentials["url"],dbName))